/**********************************************************************/
/*                                                                    */
/*                              parse.c                               */
/*                                                                    */
/*      Copyright (C) 1990 Ingo Feulner, Magic Soft Developments      */
/*                                                                    */
/*                   parsing routines for sendmail                    */
/*                                                                    */
/**********************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "config.h"
#include "log.h"
#include "addr.h"
#include "support.h"

#define LOCALMAIL       0
#define FORWARDMAIL     1
#define UNKNOWNMAIL     (-1)

#define SUCCEED         (0)
#define FAIL            (-1)

#define SYSTEM          1
#define DOMAIN          2

/* imports: */
extern char *Progname;
extern char *DefaultNode;
extern short DataBase;

/* Prototypes: */
extern datum fetch(datum);
extern char *preparse_address(char *, char **);
extern int   parse_address(char *, char **, char **);
extern int   DomainLookUp(char *, int, char *, int);

int      parse(char *, char **);
char    *split_target(char *, char **);
datum    check_domain(char *);
char    *check_domain_file(char *, char *, char *);
char    *correct_remainder(char *);

int parse(char *address, char **path)
{
  datum pathd, key;
  char *rfc_route = NULL;/* Contains copy of address if RFC_ROUTE or RFC_ENDROUTE */
  char *addr;            /* preparsed address */
  char *error;           /* possible error message */
  char *target;          /* target returned by parse_address */
  char *remainder;       /* remainder returned by parse_address */
  int form;              /* form from parse_address */


  /* First, lets check if the address contains a '!' and no '@' */
  if(strchr(address, '!') != NULL && strchr(address, '@') == NULL)
  {
    char *node = strsave(address);
    *(strchr(node, '!')) = '\0';

    /* Is this a real node name or a domain? */
    if((strchr(node, '.')) == NULL) // real node, so don't parse...
    {
      if(is_in_L_sys_file(node))
      {
        *path = strsave(address);
        free(node);
        return FORWARDMAIL;
      }
      else // Illegal address
      {
        *path = NULL;
        return UNKNOWNMAIL;
      }
    }
    free(node);
  }

  /* preparse the adress to get rid of mutant forms */
  addr = preparse_address(address, &error);
  if(addr != NULL)
  {
    char *finalpath;    /* final path we will return */
    char *node;         /* pure node */
    char *domain;       /* pure domain */

    form = parse_address(addr, &target, &remainder);
    switch(form)
    {
      case LOCAL:
        *path = remainder;
        return LOCALMAIL;
        break;

      case RFC_ROUTE:
      case RFC_ENDROUTE:
        rfc_route = strsave(address);
        break;

      case MAILBOX:
      case UUCP_ROUTE:
      case PCT_MAILBOX:
      case BERKENET:
      case DECNET:
        break;
      
      case PARSE_ERROR:
      case FAIL:
        fprintf(stderr, "%s: parse() failed: %s\n", Progname, remainder);
        *path = NULL;
        return UNKNOWNMAIL;
        break;

      default:
        fprintf(stderr, "%s: parse() returned unknown error - aborting.\n", Progname);
        exit(20);
        break;
    }

    /* Correct remainder

       a!b!cuser   -> cuser%b%a
       cuser@a     -> cuser%a
       b!c!cuser%a -> cuser%c%b%a
    */
    remainder = correct_remainder(remainder);
    
    node = split_target(target, &domain);

    if(*node != '\0')
    {
      if(DataBase != 0)
      {
        /* If there is a dummy ".uucp" domain, then cut it away */
        if(domain != NULL
          && ((stricmp(domain, ".UUCP") == 0) || (strcmp(domain, ".") == 0)) )
        {
          key.dptr  = node;
          key.dsize = strlen(node) + 1;
        }
        else
        {
          key.dptr  = target; // full address, site.domain
          key.dsize = strlen(target) + 1;
        }

        pathd = fetch(key);
      }
      else
      {
        pathd.dptr = NULL;
      }
    }
    else /* Ups, no node given! */
    {
      fprintf(stderr, "%s: no node in given address\n", Progname);
      *path = NULL;
      return UNKNOWNMAIL;
    }

    if(pathd.dptr != NULL)
    {
      if((strcmp(pathd.dptr, "%s")) != 0) // Not our system!
      {
        finalpath = xmalloc(pathd.dsize + strlen(remainder) + 1);
        sprintf(finalpath, pathd.dptr, remainder);
        *path = finalpath;

        if((strchr(finalpath, '!')) != NULL)
          return FORWARDMAIL;
        else
          return LOCALMAIL;
      }
      else // our system... so we have to check the remainder
      {
        char *newpath;
        int returncode;

        returncode = parse(remainder, &newpath);
        *path = newpath;
        return returncode;
      }
    }
    else /* Unknown node, so let's check the domainfile */
    {
      if(rfc_route != NULL) // Is this a RFC_ROUTE?
      {
        finalpath = xmalloc(strlen(DefaultNode) + strlen(rfc_route) + 2);
        sprintf(finalpath, "%s!%s", DefaultNode, rfc_route);
        *path = finalpath;
        free(rfc_route);
        rfc_route = NULL;

        return FORWARDMAIL;
      }

      // If not dummy ".uucp" domain -> search
      if(domain != NULL && (stricmp(domain, ".UUCP")) != 0)
      {
        char tmp[512]; /* Temporary Buf */

        /* Is there an entry in the Domain file for the system? */
        if(DomainLookUp(node, strlen(node), tmp, SYSTEM)) /* Yep */
        {
          finalpath = xmalloc(strlen(tmp) + strlen(remainder) + 1);
          sprintf(finalpath, tmp, remainder);
          *path = finalpath;

          if((strchr(finalpath, '!')) != NULL)
            return FORWARDMAIL;
          else
            return LOCALMAIL;
        }

        if(domain != NULL)
        {
          pathd = check_domain(domain);

          if(pathd.dptr != NULL)
          {
            finalpath = xmalloc(pathd.dsize + strlen(target) + strlen(remainder) + 2);
            sprintf(finalpath, pathd.dptr, remainder);
            strcat(finalpath, "%");
            strcat(finalpath, target);
            *path = finalpath;

            if((strchr(finalpath, '!')) != NULL)
              return FORWARDMAIL;
            else
              return LOCALMAIL;
          }
          else /* The domain is not in palias.dat */
          {    /* Last chance: Lookup in the domains file */
            finalpath = check_domain_file(domain, target, remainder);
            *path = finalpath;

            if((strchr(finalpath, '!')) != NULL)
              return FORWARDMAIL;
            else
              return LOCALMAIL;
          }
        }
      }
    }
  }
  else
  {
    fprintf(stderr, "preparse_address() returned '%s'\n", error);
  }

  *path = NULL;
  return UNKNOWNMAIL;
}

char *__regargs correct_address(char *addr)
{
  char *address = strsave(addr);
  char *ptr;

  if((strchr(address, '!')) == NULL) // No bang path
  {
    if((strchr(address, '@')) == NULL && (ptr = strrchr(address, '%')) != NULL)
    {
      *ptr = '@'; // Change last '%' into '@'
    }
  }
  return address;
}

/*
   Correct remainder

   a!b!cuser   -> cuser%b%a
   cuser@a     -> cuser%a
   cuser@c@b@a -> cuser%c%b%a
   b!c!cuser%a -> cuser%c%b%a
*/

char *correct_remainder(char *remain)
{
  char *tmpbuf;
  char *remainder;
  char *ptr;
  int length;

  tmpbuf = strsave(remain);
  length = strlen(tmpbuf);

  if((ptr = strchr(tmpbuf, '@')) != NULL)
  {
    *ptr = '%';

    if(*++ptr != '\0')
    {
      while((ptr = strchr(ptr, '@')) != NULL)
      {
        *ptr = '%';
        if(*++ptr == '\0')
          break;
      }
    }
  }

  if((strchr(tmpbuf, '!')) != NULL)
  {
    char **array;
    char *tmp;
    char *percpart = NULL;
    char *p;
    int i, n;

    tmp = xcalloc(strlen(tmpbuf) + 3);

    if((p = strchr(tmpbuf, '%')) != NULL)
    {
      percpart = strsave(p+1); // without the '%'
      *p = '\0';
    }

    strcpy(tmp, tmpbuf);
    i = strlen(tmp);

    *(tmp + i) = '\0'; // Second '\0' byte

    p = tmp;
    while((p = strchr(p, '!')) != NULL)
    {
      *p = '\0';
      if(*++p == '\0')
        break;
    }
  
    array = xcalloc(i * sizeof(char *)); // allocate array
    n = strbpl(array, i, tmp);               // build pointer list

    memset(tmpbuf, 0, length); // set tmpbuf to 0
    while(--n >= 0)
    {
      strcat(tmpbuf, array[n]);
      strcat(tmpbuf, "%");
    }

    if(percpart != NULL)
    {
      strcat(tmpbuf, percpart);
      free(percpart);
    }
    else // delete last '%'
    {
      *(strrchr(tmpbuf, '%')) ='\0';
    }

    free(tmp);
    free(array);
  }

  remainder = strsave(tmpbuf);
  free(tmpbuf);

  return remainder;
}

static char *split_target(char *target, char **domain)
{
  char *tmpt;
  char *ptr;

  tmpt = strsave(target);

  /* Includes the target node a domain? */
  if((ptr = strchr(tmpt, DOT)) != NULL)
  {
    *domain = strsave(ptr);
    *ptr = '\0';
  }
  else
  {
    *domain = NULL;
  }

  return tmpt;
}

static datum check_domain(char *domain)
{
  datum pathd, key;
  char *p;

  if(DataBase != 0) /* Only possible if maps database is available */
  {
    p = strchr(domain, DOT);
    while(p != NULL)
    {
      key.dptr  = p;
      key.dsize = strlen(p) + 1;

      pathd = fetch(key);
      if(pathd.dptr != NULL)
        break;

      if(*++p != '\0')
        p = strchr(p, DOT);
    }
  }
  else
  {
    pathd.dptr = NULL;
  }

  return pathd;
}

static char *check_domain_file(char *domain, char *target, char *remainder)
{
  char address[512];
  char *path;

  if(DomainLookUp(domain, strlen(domain), address, DOMAIN))
  {
    short i;
    char *tmp = xcalloc(strlen(target) + strlen(remainder) + 128);

    for(i = 0; i < sizeof(address) && address[i] != '!' && address[i] != '\0'; i++)
      tmp[i] = address[i];
    tmp[i] = '\0';

    if(!is_in_L_sys_file(tmp))
    {
      fprintf(stderr, "Node '%s' is not in L.sys\n", tmp);
      exit(20);
    }

    strcpy(tmp, remainder);
    strcat(tmp, "%");
    strcat(tmp, target);

    /* tmp contains "user%node.dom.ain" */

    path = xmalloc(strlen(address) + strlen(tmp) + 1);
    sprintf(path, address, tmp);

    free(tmp);
  }
  else
  {
    path = xmalloc(strlen(DefaultNode) + strlen(target) + strlen(remainder) + 3);
    sprintf(path, "%s!%s%%%s", DefaultNode, remainder, target);
  }

  return(path);
}
